#
# Copyright (c) 2015 Midokura SARL, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import subprocess
import tempfile

from charmhelpers.core import hookenv
from charmhelpers.core.services import base
import jinja2
import yaml

JUJU_HIERA_PATH = '/var/lib/hiera/'

manifest = """{{%- if fingerprint -%}}
apt::key {{ 'key':
    key => '{{{{fingerprint}}}}',
{{%- if pgp_block %}}
    key_content => '{{{{pgp_block}}}}',
{{%- endif %}}
}}

{{% endif -%}}
class {{'midonet::repository':}}
{actions}

Apt::Key <| name == 'key' |> -> Apt::Source <| |> ->
{ordering}
"""


class PuppetCallback(base.ManagerCallback):
    def __call__(self, manager, service_name, event_name):
        service = manager.get_service(service_name)
        req_data = {}
        for ctx in service.get('required_data', []):
            req_data.update(ctx)

        ctxt = {
            'fingerprint': req_data.get(self._fingerprint_key),
            'pgp_block': req_data.get(self._content_key)
        }
        ctxt.update(self._ctxt)

        with tempfile.NamedTemporaryFile(mode='w+t') as tmp:
            content = jinja2.Environment().from_string(self._manifest).render(
                **ctxt)
            tmp.write(content)
            tmp.flush()
            try:
                _apply(tmp.name)
            except subprocess.CalledProcessError:
                hookenv.log('Puppet manifest failed to apply with content %s' %
                            content)
                raise

    def __init__(self, actions, ordering, ctxt):
        self._fingerprint_key = 'midonet_key_fingerprint'
        self._content_key = 'midonet_key_content'
        self._manifest = manifest.format(actions=actions, ordering=ordering)
        self._ctxt = ctxt


class ExecutPuppetClassCallback(PuppetCallback):
    execution_template = """class {'{{component}}':}"""
    ordering_template = """Class['{{component}}']"""

    def __init__(self, component):
        super(ExecutPuppetClassCallback, self).__init__(
            actions=self.execution_template,
            ordering=self.ordering_template,
            ctxt={'component': component})


class InstallPackagesCallback(PuppetCallback):
    action_template = """package { %s:
    ensure => present
}"""

    def __init__(self, packages=()):
        actions = []
        for package in packages:
            actions.append(self.action_template % package)
        super(InstallPackagesCallback, self).__init__(
            actions='\n'.join(actions),
            ordering="""Package<| |>""",
            ctxt={})


InstallCallback = ExecutPuppetClassCallback


def _apply(path, debug=False):
    """Apply a puppet manifest."""
    cmd = ['puppet', 'apply']
    if debug:
        cmd.append('-d')
    cmd.append(path)
    subprocess.check_call(cmd)


apply = _apply


def module_install(module_name, force=False, debug=False):
    """Install or upgrade a puppet module."""
    cmd = ['puppet', 'module']
    if module_name in subprocess.check_output(['puppet', 'module', 'list']):
        cmd.append('upgrade')
    else:
        cmd.append('install')
    if force:
        cmd.append('-f')
    if debug:
        cmd.append('-d')
    cmd.append(module_name)
    subprocess.check_call(cmd)


def set_juju_hierarchy(attempts=10, extra_data_dirs=None,
                       hiera_path='/etc/puppet/hiera.yaml'):
    """Sets a puppet datadir lookup to find our customizations.

    Extra data dirs must be relative to /var/lib/hiera
    """
    if attempts < 1:
        raise RuntimeError('Failed to set juju puppet hierarchy')

    try:
        with open(hiera_path) as hiera:
            hiera_info = yaml.load(hiera)
    except IOError as ioe:
        if ioe.errno != os.errno.ENOENT:
            raise
        hiera_info = None

    if hiera_info is None:
        hiera_info = {
            ':backends': ['yaml', 'module_data'],
            ':logger': 'console',
            ':hierarchy': ['common'],
            ':yaml': {
                ':datadir': JUJU_HIERA_PATH}}

    if extra_data_dirs is not None:
        if ':hierarchy' not in hiera_info:
            hiera_info[':hierarchy'] = ['common']
        _ensure_dirs(extra_data_dirs)
        for dir_path in extra_data_dirs:
            dir_path_common = os.path.join(dir_path, 'common')
            if dir_path_common not in hiera_info[':hierarchy']:
                hiera_info[':hierarchy'].insert(0, dir_path_common)

    hiera_info[':yaml'][':datadir'] = JUJU_HIERA_PATH

    with open(hiera_path, 'w') as hiera:
        yaml.dump(hiera_info, stream=hiera, default_flow_style=False,
                  explicit_start=True, indent=2)


def _ensure_dirs(dirs):
    """Creates the directories in dirs if they are not present."""
    for dir_path in dirs:
        try:
            os.makedirs(dir_path)
        except OSError as ose:
            if ose.errno != os.errno.EEXIST:
                raise
