#
# Copyright (c) 2015 Midokura SARL, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from charmhelpers.contrib.hahelpers import cluster
from charmhelpers.contrib.network import ip
from charmhelpers.core import hookenv
from charmhelpers.core.services import helpers


class IncomingKeystoneRelation(helpers.RelationContext):
    name = 'keystone'
    interface = 'keystone'

    def __init__(self, *args, **kwargs):
        self.required_keys = ['admin_token',
                              'auth_protocol',
                              'service_protocol',
                              'service_host',
                              'service_port',
                              'auth_host',
                              'auth_port',
                              'service_username',
                              'service_password',
                              'service_tenant',
                              'service_tenant_id']
        helpers.RelationContext.__init__(self, *args, **kwargs)


class MEMManagerRelation(helpers.RelationContext):
    """Midokura Enterprise Manager relation

    Optional relation to set the base_uri of the MidoNet API to match that
    of the reverse proxy in MEM Manager
    """
    name = 'manager'
    interface = 'midonet-manager'

    def __init__(self, *args, **kwargs):
        self.required_keys = ['base_uri']
        helpers.RelationContext.__init__(self, *args, **kwargs)


class OutgoingKeystoneRelation(helpers.RelationContext):
    name = 'keystone'
    interface = 'keystone'

    def __init__(self, admin_url, internal_url, public_url, region, service,
                 *args, **kwargs):
        self.required_keys = ['admin_url',
                              'internal_url',
                              'public_url',
                              'region',
                              'service']
        self.admin_url = admin_url
        self.internal_url = internal_url
        self.public_url = public_url
        self.region = region
        self.service = service
        helpers.RelationContext.__init__(self, *args, **kwargs)

    def provide_data(self):
        return {
            'admin_url': self.admin_url,
            'internal_url': self.internal_url,
            'public_url': self.public_url,
            'region': self.region,
            'service': self.service
        }


class HARelation(helpers.RelationContext):
    name = 'ha'
    interface = 'hacluster'

    def __init__(self, *args, **kwargs):
        helpers.RelationContext.__init__(self, *args, **kwargs)
        self.required_keys = ['init_services',
                              'corosync_bindiface',
                              'corosync_mcastport',
                              'resources',
                              'resource_params',
                              'clones']

        cluster_config = cluster.get_hacluster_config()
        self.iface = cluster_config['ha-bindiface']
        self.mcastport = cluster_config['ha-mcastport']
        self.resources = None
        self.resource_params = None
        resources = {'res_midonet_tomcat7': 'lsb:tomcat7'}
        resource_params = {'res_midonet_tomcat7': 'op monitor interval="5s"'}

        vip = cluster_config['vip']
        if not vip:
            return

        iface = (ip.get_iface_for_address(vip) or hookenv.config('vip_iface'))
        netmask = (ip.get_netmask_for_address(vip) or
                   hookenv.config('vip_cidr'))

        if iface is not None:
            key = 'res_midonet_%s_vip' % iface
            resources[key] = 'ocf:heartbeat:IPaddr2'
            resource_params[key] = ('params ip="%s" cidr_netmask="%s" nic="%s"'
                                    % (vip, netmask, iface))
            self.resources = resources
            self.resource_params = resource_params

    def provide_data(self):
        return {
            'init_services': {'res_midonet_tomcat7': 'tomcat7'},
            'corosync_bindiface': self.iface,
            'corosync_mcastport': self.mcastport,
            'resources': self.resources,
            'resource_params': self.resource_params,
            'clones': {'cl_midonet_tomcat7': 'res_midonet_tomcat7'}}
