#
# Copyright (c) 2015 Midokura SARL, All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from charmhelpers.contrib.hahelpers import cluster
from charmhelpers.contrib.network import ip
from charmhelpers.contrib.openstack import utils as helpers_utils
from charmhelpers.core import decorators
from charmhelpers.core import hookenv
from charmhelpers.core.services import base
from charmhelpers.core.services import helpers
from keystoneclient.v2_0 import client
from midonet_helpers import puppet
from midonet_helpers import relations as common_relations
from midonet_helpers import repositories
import relations
import utils
from webob import exc

CREDS_PATH = '/etc/midonet_creds'
CLUSTER_RES_GRP_NAME = 'grp_midonet_vips'


class ApiConfig(dict):
    def __init__(self, *args):
        self['config'] = hookenv.config()
        self['config']['api_ip'] = utils.get_ip()


class TunnelHandling(base.ManagerCallback):
    def __init__(self):
        retrier = decorators.retry_on_exception(num_retries=10, base_delay=1,
                                                exc_type=exc.HTTPError)
        self._default_tunnel_zone = retrier(self._get_default_tunnel_zone)

    def __call__(self, manager, service_name, event_name):
        if not cluster.is_elected_leader(CLUSTER_RES_GRP_NAME):
            hookenv.log('This unit is not the leader, deferring tunnel zone '
                        'handling to the midonet-api cluster leader.')
            return
        service = manager.get_service(service_name)
        context = {}
        for ctx in service.get('required_data', []):
            context.update(ctx)
        self._add_to_tunnel_zone(context)

    def _get_default_tunnel_zone(self, conn, tunnel_type='vxlan',
                                 name='default_tz'):
        """Returns the default tunnel zone. Creates it if it did not exist."""
        tunnel_zones = [tz for tz in conn.get_tunnel_zones() if
                        tz.get_name() == name]
        if tunnel_zones:
            tz = tunnel_zones[0]
        else:
            ntz = conn.add_tunnel_zone()
            ntz.type(tunnel_type)
            ntz.name(name)
            try:
                ntz.create()
            except Exception as ex:
                hookenv.log('Failed to create the default tunnel zone. Err: %s'
                            % ex, level=hookenv.ERROR)
            tz = conn.get_tunnel_zone(ntz.get_id())
        return tz

    def _get_midonet_host_uuids(self):
        """Generates an iterable of related ids to add to the tunnel zone."""
        for rid in hookenv.relation_ids('host'):
            for unit in hookenv.related_units(rid):
                host_uuid = hookenv.relation_get('host_uuid', unit=unit,
                                                 rid=rid)
                if host_uuid is None:
                    hookenv.log('No uuid received for a host relation.',
                                level=hookenv.WARNING)
                else:
                    yield host_uuid

    def _get_keystone_token(self, context):
        keystone_client = client.Client(
            username=context['keystone'][0]['service_username'],
            password=context['keystone'][0]['service_password'],
            tenant_name=context['keystone'][0]['service_tenant'],
            auth_url='http://%s:%s/v2.0' %
            (context['keystone'][0]['auth_host'],
             context['keystone'][0]['auth_port']))
        return keystone_client.get_token(keystone_client.session)

    def _get_midonet_connection(self, ctxt):
        """Returns an open connection object to MidoNet API."""
        from midonetclient import api
        api.auth_lib._token = self._get_keystone_token(ctxt)
        login_data = {
            'base_uri': 'http://%s:%s/midonet-api' %
                        (utils.get_ip(), ctxt['config']['api_port']),
            'username': ctxt['keystone'][0]['service_username'],
            'password': ctxt['keystone'][0]['service_password'],
            'project_id': ctxt['keystone'][0]['service_tenant']}
        try:
            conn = api.MidonetApi(**login_data)
        except Exception:
            hookenv.log('Failed to authenticate to midonet-api, with login '
                        'data %r. It may be stopped will retry on the next '
                        'event.' % login_data, level=hookenv.WARNING)
            raise
        return conn

    def _get_tz_host_uuids(self, tunnel_zone):
        return frozenset(host.get_host_id() for
                         host in tunnel_zone.get_hosts())

    def _add_to_tunnel_zone(self, context):
        """Adds related hosts to the default tunnel zone."""
        host_uuids = list(self._get_midonet_host_uuids())
        if not host_uuids:
            hookenv.log('No hosts to add yet.', level=hookenv.INFO)
            return

        conn = self._get_midonet_connection(context)
        tz = self._default_tunnel_zone(conn)
        tz_host_uuid_set = self._get_tz_host_uuids(tz)

        agent_hosts_map = dict((host.get_id(), host) for
                               host in conn.get_hosts())
        for uuid in list(host_uuids):
            if uuid not in tz_host_uuid_set:
                host = agent_hosts_map.get(uuid)
                if host is None:
                    hookenv.log('Host %s not found in midonet agent hosts %r' %
                                (uuid, agent_hosts_map), level=hookenv.INFO)
                    continue
                agent_hostname = host.get_name()
                new_tz_host = tz.add_tunnel_zone_host()
                new_tz_host.host_id(uuid)
                addr = ip.get_host_ip(agent_hostname)
                new_tz_host.ip_address(addr)
                try:
                    new_tz_host.create()
                except Exception as ex:
                    hookenv.log(
                        'Failed to add host %s(%s) with address %s to the '
                        'default tunnel zone. Err: %s' % (agent_hostname, uuid,
                                                          addr, ex),
                        level=hookenv.ERROR)
                    continue
                else:
                    hookenv.log(
                        'host %s(%s) with address: %s added to the tunnel '
                        'zone' % (agent_hostname, uuid, addr),
                        level=hookenv.INFO)


class MidonetApiRelation(helpers.RelationContext):
    name = 'midonet-api'
    interface = 'midonet'
    port = 8080

    def __init__(self, *args, **kwargs):
        self.required_keys = []
        helpers.RelationContext.__init__(self, *args, **kwargs)

    def provide_data(self):
        return {
            'host': utils.get_ip(),
            'port': self.port,
        }


def manage():
    midonet_api_relation = MidonetApiRelation()
    cfg = ApiConfig()
    ost_origin = cfg['config'].get('openstack-origin')
    midonet_origin = cfg['config'].get('midonet-origin')
    mem_username = cfg['config'].get('mem-username')
    mem_pwd = cfg['config'].get('mem-password')
    if None in (ost_origin, midonet_origin) or (
            midonet_origin is not None and midonet_origin[:3] == 'mem' and None
            in (mem_username, mem_pwd)):
        hookenv.status_set('maintenance', 'lacking repo configuration')
        return

    ost_release = helpers_utils.get_os_codename_install_source(ost_origin)
    repo = repositories.config('%s/%s' % (ost_release, midonet_origin),
                               mem_username, mem_pwd)
    mem = relations.MEMManagerRelation()
    out_keystone = relations.OutgoingKeystoneRelation(
        service='midonet', region=hookenv.config('region'), **utils.urls())

    manager = base.ServiceManager([
        {
            'service': 'tomcat7',
            'ports': [midonet_api_relation.port],
            'provided_data': [midonet_api_relation,
                              out_keystone],
            'required_data': [cfg,
                              relations.IncomingKeystoneRelation(),
                              common_relations.ZookeeperRelation(),
                              repo],
            'data_ready': [
                helpers.render_template(
                    source='common.yaml',
                    target='/var/lib/hiera/midonet_api/common.yaml'),
                puppet.PuppetCallback(
                    actions="""class {'midonet::midonet_api':}""",
                    ordering="""Class['midonet::midonet_api::install']""",
                    ctxt={}),
                puppet.InstallCallback('midonet::midonet_cli'),
                helpers.render_template(
                    source='tomcat7',
                    target='/etc/default/tomcat7'),
                helpers.render_template(
                    source='.midonetrc',
                    target='/root/.midonetrc'),
                helpers.render_template(
                    source='midonet-cli',
                    target='/usr/sbin/midonet-cli',
                    perms=0o777),  # midonet-cli workaround
            ],
            'start': [base.service_restart,
                      base.PortManagerCallback(),
                      TunnelHandling()],
        },
    ])

    # ServiceManagers detect empty relations with no required keys as
    # incomplete. Thus, those relations that are entirely optional, should
    # only be added to 'required_data' if they bring data.
    if mem:
        manager.get_service('tomcat7').get('required_data', []).append(
            mem)

    manager.manage()
