#!/usr/bin/env python3
# vim: tabstop=4 shiftwidth=4 softtabstop=4 filetype=python
#
# Copyright (c) 2015 Midokura Europe SARL, All Rights Reserved.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import uuid

import amulet

from charmhelpers.contrib.openstack.amulet.utils import (
    OpenStackAmuletUtils,
    DEBUG,  # flake8: noqa
    ERROR
)

# Use DEBUG to turn on debug logging
u = OpenStackAmuletUtils(DEBUG)

SETUP_TIMEOUT = 9600


class ApiMinimumDeployment():
    def __init__(self, ubuntu_series, openstack_origin=None,
                 midonet_origin=None):
        self.os_release = openstack_origin
        self.series = ubuntu_series
        self.midonet_release = (openstack_origin.split('-')[-1] + '/' +
            midonet_origin)
        self.midonet_origin = midonet_origin
        self.mem = self.midonet_origin[:3] == 'mem'
        self.mem_credentials = {'mem-username': os.getenv('MEM_USERNAME'),
                                'mem-password': os.getenv('MEM_PWD')}
        self.d = amulet.Deployment(series=self.series)
        self.d.charm_name = 'midonet-api'

        # Sentries
        self.zookeeper_sentry = None
        self.keystone_sentry = None
        self.cassandra_sentry = None
        self.nova_compute_sentry = None
        self.mn_api_sentry = None
        self.keystone_api_relation = None
        self.zookeeper_api_relation = None

        self._add_services()
        self._add_relations()
        self._configure_services()
        self._deploy()
        self._initialize_tests()

    def _add_services(self):
        self.d.add('mysql', charm='mysql', units=1,
                   series=self.series)
        self.d.add('keystone', charm='keystone', units=1,
                   series=self.series)
        self.d.add('zookeeper', charm='zookeeper', units=1,
                   series=self.series)
        self.d.add('cassandra', charm='cassandra',
                   units=1, series=self.series)
        self.d.add('compute', charm='nova-compute', units=1,
                   series=self.series, constraints={'mem': "4G"})
        self.d.add('midonet-agent', charm='midonet-agent',
                   branch='lp:~celebdor/charms/trusty/midonet-agent/trunk',
                   series=self.series)
        self.d.add('midonet-api')

    def _add_relations(self):
        self.d.relate('mysql:shared-db', 'keystone:shared-db')
        self.d.relate('keystone:identity-service', 'midonet-api:keystone')
        self.d.relate('zookeeper:zookeeper', 'midonet-api:zookeeper')

        self.d.relate('midonet-agent:cassandra', 'cassandra:database')
        self.d.relate('midonet-agent:zookeeper', 'zookeeper:zookeeper')

        self.d.relate('midonet-agent:neutron-plugin', 'compute:neutron-plugin')
        self.d.relate('midonet-agent:host', 'midonet-api:host')

    def _configure_services(self):
        self.d.configure('keystone', {
                         'admin-user': 'admin',
                         'admin-password': 'openstack',
                         'admin-token': 'osttesting',
                         'enable-pki': 'false',
                         'openstack-origin': self.os_release})

        self.d.configure('cassandra', {
            'authenticator': 'AllowAllAuthenticator',
            'cluster_name': 'midonet',
            'install_keys': '7E41C00F85BFC1706C4FFFB3350200F2B999A372',
            'install_sources':
            'deb http://debian.datastax.com/community 2.0 main',
            'extra_packages': 'dsc20'})

        self.d.configure('mysql', {'max-connections': 500})

        api_config = {'midonet-origin': self.midonet_origin,
                      'openstack-origin': self.os_release}
        if self.mem:
            api_config.update(self.mem_credentials)
        self.d.configure('midonet-api', api_config)
        self.d.configure('midonet-agent', api_config)


    def _deploy(self):
        try:
            self.d.setup(timeout=SETUP_TIMEOUT)
            self.d.sentry.wait(timeout=SETUP_TIMEOUT)
        except amulet.helpers.TimeoutError:
            amulet.raise_status(amulet.SKIP,
                                msg="Environment wasn't stood up in time")

    def _initialize_tests(self):
        self.zookeeper_sentry = self.d.sentry.unit['zookeeper/0']
        self.keystone_sentry = self.d.sentry.unit['keystone/0']
        self.cassandra_sentry = self.d.sentry.unit['cassandra/0']
        self.mn_api_sentry = self.d.sentry.unit['midonet-api/0']

        self.keystone_api_relation = self.keystone_sentry.relation(
            'identity-service', 'midonet-api:keystone')
        self.zookeeper_api_relation = self.zookeeper_sentry.relation(
            'zookeeper', 'midonet-api:zookeeper')

    def run_tests(self):
        for test in dir(self):
            if test.startswith('test_'):
                getattr(self, test)()

    def test_mn_api(self):
        u.log.debug('Checking if the repo key has been correctly fetched...')
        apt_key_out, _ = self.mn_api_sentry.run(
            'apt-key list | grep Midokura')
        if 'Midokura' not in apt_key_out:
            message = ("error, Midokura repository key is not installed")
            amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking if midonet-api package has been installed...')
        dpkg_api_out, _ = self.mn_api_sentry.run(
            'dpkg -l | grep midonet-api')
        if 'midonet-api' not in dpkg_api_out:
            message = ("error: midonet-api package is not installed")
            amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking if midonetclient package has been installed...')
        dpkg_client_out, _ = self.mn_api_sentry.run(
            'dpkg -l | grep midonetclient')
        if 'python-midonetclient' not in dpkg_client_out:
            message = ("error: python-midonetclient package is not installed")
            amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking if tomcat7 is installed and running...')
        tomcat7_out, tomcat7_exit = self.mn_api_sentry.run(
                                    'service tomcat7 status')
        if tomcat7_exit == 1:
            message = ("midonet-api setup error, tomcat7 is not installed")
            amulet.raise_status(amulet.FAIL, msg=message)
        if tomcat7_exit == 0:
            if 'running' not in tomcat7_out:
                message = ("midonet-api setup error, tomcat7 is not running")
                amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking if midonetrc exists...')
        mn_rc_out, _ = self.mn_api_sentry.run(
            '[ -f /root/.midonetrc ] && echo \"True\" || echo \"False\"')
        if 'True' not in mn_rc_out:
            message = ("error: .midonetrc not found")
            amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking midonet-cli...')
        cli_out, mn_exit = self.mn_api_sentry.run(
                           'midonet-cli -e create bridge name test')
        if mn_exit != 0:
            message = ("midonet-cli error: %s" % cli_out)
            amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking if the bridge has been successfully created...')
        cli_out, mn_exit = self.mn_api_sentry.run('midonet-cli -e bridge list')
        if 'test' not in cli_out:
            message = ("midonet-cli error: %s" % cli_out)
            amulet.raise_status(amulet.FAIL, msg=message)

        u.log.debug('Checking if the host has been added to the tunnel '
                    'zone...')
        cli_out, mn_exit = self.mn_api_sentry.run(
            'midonet-cli -e tunnel-zone list')
        if cli_out:
            tz_uuid = None
            try:
                tz_uuid = uuid.UUID(cli_out.split(' ')[1], version=4)
            except Exception as ex:
                amulet.raise_status(
                    amulet.FAIL,
                    msg='Failed to verify tunnel zone creation and membership.'
                        ' %s' % ex)

            cli_out, mn_exit = self.mn_api_sentry.run(
                'midonet-cli -e tunnel-zone %s member list' % tz_uuid)
            if 'host' not in cli_out:
                message = ("Failed to see host in the tunnel zone: %s" % cli_out)
                amulet.raise_status(amulet.FAIL, msg=message)
        else:
            message = ("Failed to detect tunnel-zone uuid")
            amulet.raise_status(amulet.FAIL, msg=message)

