#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 filetype=python
#
# Copyright (c) 2015 Canonical Ltd.
# Copyright (c) 2015 Midokura Sarl.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Taken from the neutron-api charm. It should be included in charm-helpers
# in the future. (It was under unit_tests/test_utils.py).
import contextlib
import io
import logging
import os
import unittest

import mock
import yaml


def load_config():
    """Get config options

    Walk backwords from __file__ looking for config.yaml, load and return the
    'options' section'
    """
    config = None
    directory = os.path.dirname(__file__)
    while config is None:
        if os.path.isfile(os.path.join(directory, 'config.yaml')):
            config = os.path.join(directory, 'config.yaml')
            break
        parent_dir = os.path.dirname(directory)
        if parent_dir == directory:
            break  # We reached the root of the fs
        directory = parent_dir

    if not config:
        logging.info('Could not find config.yaml in any parent directory '
                     'of %s. ' % __file__)
        return

    return yaml.safe_load(open(config).read())['options']


class ServiceConfig(dict):
    def __init__(self):
        self.implicit_save = False
        dict.__init__(self)


def get_default_config():
    """Get the default charm config

    Load default charm config from config.yaml return as a dict.
    If no default is set in config.yaml, its value is None.
    """
    default_config = ServiceConfig()
    config = load_config()
    if config is not None:
        for k, v in config.items():
            if 'default' in v:
                default_config[k] = v['default']
            else:
                default_config[k] = None
    return default_config


class CharmTestCase(unittest.TestCase):

    def setUp(self, obj, patches):
        super(CharmTestCase, self).setUp()
        self._patches = patches
        self._obj = obj
        self.test_config = TestConfig()
        self.test_relation = TestRelation()
        self.patch_all(self._obj, self._patches)

    def patch(self, obj, method):
        _m = mock.patch.object(obj, method)
        _mock = _m.start()
        self.addCleanup(_m.stop)
        return _mock

    def patch_all(self, obj, patches):
        for method in patches:
            setattr(self, method, self.patch(obj, method))


class TestConfig(object):

    def __init__(self):
        self.config = get_default_config()

    def get(self, attr=None):
        if not attr:
            return self.get_all()
        try:
            return self.config[attr]
        except KeyError:
            return None

    def get_all(self):
        return self.config

    def set(self, attr, value):
        """Allows to set even non-original values to match hookenv.Config."""
        self.config[attr] = value


class TestRelation(object):

    def __init__(self, relation_data={}):
        self.relation_data = relation_data

    def set(self, relation_data):
        self.relation_data = relation_data

    def get(self, attribute=None, unit=None, rid=None):
        if attribute is None:
            return self.relation_data
        elif attribute in self.relation_data:
            return self.relation_data[attribute]
        return None


class FakeRelation(object):
    """A fake relation class.

    Lets tests specify simple relation data
    for a default relation + unit (foo:0, foo/0, set in setUp()), eg:

        rel = {
            'private-address': 'foo',
            'password': 'passwd',
        }
        relation = FakeRelation(rel)
        self.relation_get.side_effect = relation.get
        passwd = self.relation_get('password')

    or more complex relations meant to be addressed by explicit relation id
    + unit id combos:

        rel = {
            'mysql:0': {
                'mysql/0': {
                    'private-address': 'foo',
                    'password': 'passwd',
                }
            }
        }
        relation = FakeRelation(rel)
        self.relation_get.side_affect = relation.get
        passwd = self.relation_get('password', rid='mysql:0', unit='mysql/0')
    """
    def __init__(self, relation_data):
        self.relation_data = relation_data

    def get(self, attr=None, unit=None, rid=None):
        if not rid or rid == 'foo:0':
            if attr is None:
                return self.relation_data
            elif attr in self.relation_data:
                return self.relation_data[attr]
            return None
        else:
            if rid not in self.relation_data:
                return None
            try:
                relation = self.relation_data[rid][unit]
            except KeyError:
                return None
            if attr is None:
                return relation
            elif attr in relation:
                return relation[attr]
            return None

    def relation_ids(self, relation=None):
        if relation is None:
            return self.relation_data.keys()
        rids = []
        for key in self.relation_data.keys():
            if key.startswith(relation) and key[len(relation)] == ':':
                rids.append(key)
        return rids

    def related_units(self, relid=None):
        try:
            return list(self.relation_data[relid].keys())
        except KeyError:
            return []

    def relation_units(self, relation_id):
        if relation_id not in self.relation_data:
            return None
        return list(self.relation_data[relation_id].keys())


@contextlib.contextmanager
def patch_open():
    """Mock open()

    Mocking allows to mock both open() itself and the file that is being
    yielded.

    Yields the mock for "open" and "file", respectively.
    """
    mock_open = mock.MagicMock(spec=open)
    mock_file = mock.MagicMock(spec=io.FileIO)

    @contextlib.contextmanager
    def stub_open(*args, **kwargs):
        mock_open(*args, **kwargs)
        yield mock_file

    with mock.patch('__builtin__.open', stub_open):
        yield mock_open, mock_file
