#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 filetype=python
#
# Copyright (c) 2015 Midokura Europe SARL, All Rights Reserved.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import inspect
import shutil
import tempfile
import uuid

import mock
import services

from .midonet_helpers import test

HOOKENV_PATCHES = [
    'charm_dir',
    'config',
    'is_relation_made',
    'log',
    'open_port',
    'relation_get',
    'relation_ids',
    'related_units',
    'relation_set',
    'unit_get']


class NewTZHostMock(object):
    def __init__(self, tz):
        self._tz = tz

    def host_id(self, tz_host_uuid):
        self._host_uuid = tz_host_uuid

    def ip_address(self, addr):
        self._addr = addr

    def create(self):
        self._uuid = str(uuid.uuid4())
        self._tz._hosts.append(TZHostMock(self._uuid, self._host_uuid,
                                          self._addr))


class TZHostMock(object):
    def __init__(self, tz_host_uuid, host_uuid, addr):
        self._uuid = tz_host_uuid
        self._host_uuid = host_uuid

    def get_id(self):
        return self._uuid

    def get_host_id(self):
        return self._host_uuid


class HostMock(object):
    def __init__(self, host_uuid, name):
        self._uuid = host_uuid
        self._name = name

    def get_id(self):
        return self._uuid

    def get_name(self):
        return self._name


class NewTZMock(object):
    def __init__(self, conn):
        self._conn = conn

    def type(self, kind):
        self._kind = kind

    def name(self, name):
        self._name = name

    def create(self):
        self._uuid = str(uuid.uuid4())
        self._conn._tzs.append(TZMock(self._uuid, self._kind, self._name, []))

    def get_id(self):
        return self._uuid


class TZMock(object):
    def __init__(self, tz_uuid, kind, name, hosts):
        self._uuid = tz_uuid
        self._kind = kind
        self._name = name
        self._hosts = hosts

    def get_id(self):
        return self._uuid

    def get_name(self):
        return self._name

    def get_type(self):
        return self._kind

    def get_hosts(self):
        return self._hosts

    def add_tunnel_zone_host(self):
        return NewTZHostMock(self)


class MidonetConnMock(object):
    def __init__(self, tunnel_zones, hosts):
        self._tzs = tunnel_zones
        self._hosts = hosts

    def get_tunnel_zones(self):
        return self._tzs

    def get_hosts(self):
        return self._hosts

    def get_tunnel_zone(self, tz_uuid):
        for tz in self._tzs:
            if tz.get_id() == tz_uuid:
                return tz

    def add_tunnel_zone(self):
        return NewTZMock(self)


class HookTests(test.CharmTestCase):
    @mock.patch('charmhelpers.fetch.apt_install')
    def setUp(self, apt_install):
        self.relations = test.FakeRelation(self._full_relation_data())
        test.CharmTestCase.setUp(self, services.hookenv, HOOKENV_PATCHES)

        def lsb_release():
            ubuntu_trusty = {
                'DISTRIB_ID': 'Ubuntu',
                'DISTRIB_RELEASE': '14.04',
                'DISTRIB_CODENAME': 'trusty',
                'DISTRIB_DESCRIPTION': "Ubuntu 14.04.3 LTS",
            }

            try:
                data = services.host.lsb_release()
            except Exception:
                return ubuntu_trusty
            if data.get('DISTRIB_ID') != 'Ubuntu':
                return ubuntu_trusty
            return data

        def get_os_codename_package(pkg, fatal=True):
            version = 'liberty'
            try:
                ret_vers = services.helpers_utils.get_os_codename_package(
                    pkg, fatal)
            except Exception:
                return version
            return ret_vers

        self.patch_all(services.helpers_utils,
                       ('get_os_codename_package',
                        'lsb_release',))
        self.lsb_release.side_effect = lsb_release
        self.get_os_codename_package.side_effect = get_os_codename_package

        self.patch_all(services.utils.cluster,
                       ('is_clustered',
                        'get_hacluster_config',))
        self.is_clustered.return_value = False

        _templating = mock.patch.object(services.helpers, 'render_template')
        setattr(self, 'render_template', _templating.start())
        self.addCleanup(_templating.stop)

        _check_call = mock.patch.object(services.hookenv.subprocess,
                                        'check_call')
        setattr(self, 'check_call', _check_call.start())
        self.addCleanup(_check_call.stop)

        _puppet = mock.patch.object(services.puppet, '_apply')
        setattr(self, 'apply', _puppet.start())
        self.addCleanup(_puppet.stop)

        _cluster = mock.patch.object(services.cluster, 'is_elected_leader')
        setattr(self, 'is_elected_leader', _cluster.start())
        self.addCleanup(_cluster.stop)

        _relation_helpers = mock.patch.object(
            services.relations.helpers, 'hookenv', spec=[
                'DEBUG',
                'log',
                'relation_ids',
                'related_units',
                'relation_get'])
        setattr(self, 'relation_helpers', _relation_helpers.start())
        self.addCleanup(_relation_helpers.stop)

        self.test_config.config['openstack-origin'] = 'cloud:trusty-liberty'
        self.config.side_effect = self.test_config.get
        self.relation_get.side_effect = self.test_relation.get
        self.test_config.set('api_ip', 'dom.example.com')

        self.charm_dir.return_value = tempfile.mkdtemp()
        self.related_units.side_effect = self.relations.related_units
        self.relation_ids.side_effect = self.relations.relation_ids
        self.relation_get.side_effect = self.relations.get

        __get_midonet_connection = mock.patch.object(services.TunnelHandling,
                                                     '_get_midonet_connection')
        setattr(self, '_get_midonet_connection',
                __get_midonet_connection.start())
        self.addCleanup(__get_midonet_connection.stop)

    def tearDown(self):
        shutil.rmtree(self.charm_dir.return_value)
        test.CharmTestCase.tearDown(self)

    @mock.patch('services.ip.get_host_ip')
    @mock.patch('charmhelpers.core.services.base.service_restart')
    @mock.patch('charmhelpers.core.services.base.service_stop')
    def test_config_change(self, service_stop, service_restart, get_host_ip):
        self.relation_helpers.relation_ids.side_effect = (
            self.relations.relation_ids)
        self.relation_helpers.related_units.side_effect = (
            self.relations.related_units)
        self.relation_helpers.relation_get.side_effect = (
            self.relations.get)
        self.is_elected_leader.return_value = True

        spec = inspect.getargspec(
            services.TunnelHandling._get_default_tunnel_zone)
        defaults = dict(zip(spec.args[-len(spec.defaults):], spec.defaults))
        api_mock = MidonetConnMock(
            tunnel_zones=[
                TZMock(
                    tz_uuid='5276d75e-74c6-4d13-9007-ac0147e449cb',
                    kind=defaults['tunnel_type'],
                    name=defaults['name'],
                    hosts=[
                        TZHostMock(
                            tz_host_uuid='77a703b9-98bc-45eb-b36c-'
                                         '47636aa79759',
                            host_uuid='fd469700-5d6c-4ed7-a404-bb07bace535f',
                            addr='192.168.11.11'),
                        ]
                    ),
                ],
            hosts=[
                HostMock(
                    host_uuid='fd469700-5d6c-4ed7-a404-bb07bace535f',
                    name='foo.maas'),
                HostMock(
                    host_uuid='024d703d-efbf-4716-ba45-eb0a15842b42',
                    name='bar.maas'),
                ]
            )
        self._get_midonet_connection.return_value = api_mock

        def _get_host_ip(hostname):
            if hostname == 'bar.maas':
                return '192.168.11.22'
            raise ValueError(hostname)

        get_host_ip.side_effect = _get_host_ip

        services.manage()

        self.assertTrue(self.render_template.called)
        self.assertTrue(self.apply.called)
        self.assertTrue(service_restart.called)
        self.assertEqual(
            api_mock.get_tunnel_zones()[0].get_hosts()[1].get_host_id(),
            '024d703d-efbf-4716-ba45-eb0a15842b42')

    def _full_relation_data(self):
        return {
            'keystone:4': {
                'keystone/0': {
                    'admin_token': 'thepass',
                    'auth_protocol': 'http',
                    'service_protocol': 'http',
                    'service_host': 'keystone_h',
                    'service_port': 5000,
                    'auth_host': 'keystone_h',
                    'auth_port': 35357,
                    'service_username': 'admin',
                    'service_password': 'mypass',
                    'service_tenant': 'mytenant',
                    'service_tenant_id': 'bdbd68fb-f6e9-4143-bf1a-797494d3f2ed'
                },
            },
            'zookeeper:6': {
                'zookeeper/0': {
                    'private-address': 'zoo1',
                    'port': 2181
                },
                'zookeeper/1': {
                    'private-address': 'zoo2',
                    'port': 2181
                },
            },
            'host:3': {
                'midonet-agent/0': {
                    'private-address': 'foo.maas',
                    'public-address': 'foo.maas',
                    'host_uuid': 'fd469700-5d6c-4ed7-a404-bb07bace535f'
                },
                'midonet-agent/1': {
                    'private-address': 'bar.maas',
                    'public-address': 'bar.maas',
                    'host_uuid': '024d703d-efbf-4716-ba45-eb0a15842b42'
                }
            }
        }

    def _relation_get_full(self, rid=None, unit=None):
        return self.data[rid][unit]
