#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4 filetype=python
#
# Copyright (c) 2015 Midokura Europe SARL, All Rights Reserved.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import tempfile
import unittest

from charmhelpers.core import templating
import mock
import pkg_resources
import yaml

TEMPLATES_DIR = pkg_resources.resource_filename(__name__, '../templates')


class TestTemplates(unittest.TestCase):
    def setUp(self):
        self.name = 'myseVkrver'
        self.keystone = ('192.168.11.2', 35357)
        self.zookeeper = [
            ('192.168.11.3', 8080),
            ('192.168.11.7', 8080),
        ]
        self.context = {
            'config': {
                'api_ip': '192.168.11.11',
                'api_port': 11111,
                'vtep': True},
            'keystone': [{
                'admin_token': '123412347123435123sdaaf',
                'auth_host': self.keystone[0],
                'auth_port': self.keystone[1]}],
            'midonet_repo': 'http://repo.midonet.org/midonet/v2014.11',
            'midonet_openstack_repo': 'http://repo.midonet.org/openstack',
            'midonet_thirdparty_repo': 'http://repo.midonet.org/misc',
            'midonet_key_fingerprint': '50F18FCB',
            'midonet_stage': 'stable',
            'openstack_release': 'juno',
            'zookeeper': [{
                'private-address': host[0],
                'port': host[1]} for host in self.zookeeper],
        }

        self.charm_dir = pkg_resources.resource_filename(__name__, '')
        self._charm_dir_patch = mock.patch.object(templating.hookenv,
                                                  'charm_dir')
        self._charm_dir_mock = self._charm_dir_patch.start()
        self._charm_dir_mock.side_effect = lambda: self.charm_dir

    def tearDown(self):
        self._charm_dir_patch.stop()

    @mock.patch.object(templating.host.os, 'fchown')
    @mock.patch.object(templating.host, 'mkdir')
    @mock.patch.object(templating.host, 'log')
    def test_render_common_yaml(self, log, mkdir, fchown):
        _, temp_path = tempfile.mkstemp()
        try:
            templating.render('common.yaml', temp_path,
                              self.context, templates_dir=TEMPLATES_DIR)
            with open(temp_path) as cli_file:
                data = yaml.load(cli_file)

            self.assertEqual(data['midonet::midonet_api::api_ip'],
                             self.context['config']['api_ip'])
            # self.assertEqual(data['midonet::midonet_api::api_port'],
            # self.context['config']['api_port'])
            self.assertEqual(len(data['midonet::midonet_api::zk_servers']),
                             2)
            self.assertEqual(data['midonet::midonet_api::zk_servers'][0],
                             {'ip': self.zookeeper[0][0],
                              'port': self.zookeeper[0][1]})
            self.assertTrue(data['midonet::midonet_api::keystone_auth'])
            self.assertEqual(data['midonet::midonet_api::keystone_host'],
                             self.keystone[0])
            self.assertEqual(data['midonet::midonet_api::keystone_port'],
                             self.keystone[1])
            self.assertEqual(
                data['midonet::midonet_api::keystone_admin_token'],
                self.context['keystone'][0]['admin_token'])

            self._check_repo(data)
        finally:
            if os.path.exists(temp_path):
                os.remove(temp_path)

    def _check_repo(self, data):
        self.assertEqual(data['midonet::repository::midonet_repo'],
                         self.context['midonet_repo'])
        self.assertEqual(
            data['midonet::repository::midonet_openstack_repo'],
            self.context['midonet_openstack_repo'])
        self.assertEqual(
            data['midonet::repository::midonet_thirdparty_repo'],
            self.context['midonet_thirdparty_repo'])
        self.assertEqual(data['midonet::repository::midonet_stage'],
                         self.context['midonet_stage'])
        self.assertEqual(data['midonet::repository::openstack_release'],
                         self.context['openstack_release'])
